@@ -30,13 +30,20 @@
 import numpy as np
 import pandas as pd
 from typing import Optional, Union
+from scipy.optimize import differential_evolution
 
 # system imports
 import os
+import sys
+import glob
+import time as t
 from itertools import cycle
 from copy import deepcopy
 
-from pyrates.utility.grid_search import grid_search, ClusterGridSearch, linearize_grid
+# pyrates imports
+from pyrates.utility.grid_search import grid_search, ClusterGridSearch, linearize_grid, adapt_circuit, ClusterCompute, \
+    StreamTee
+from pyrates.frontend import CircuitTemplate
 
 # meta infos
 __author__ = "Christoph Salomon, Richard Gast"
@@ -145,6 +152,7 @@ class GeneticAlgorithmTemplate:
         iter_count = 0
         stagnation_count = 0
         while iter_count < max_iter:
+
             print("")
             print(f'ITERATION: {iter_count}')
 
@@ -174,7 +182,7 @@ class GeneticAlgorithmTemplate:
                     old_fitness = np.round(float(self.candidate['fitness']), decimals=stagnation_decimals)
                     new_fitness = np.round(self.current_max_fitness, decimals=stagnation_decimals)
                     # Check for change in fitness
-                    if new_fitness == old_fitness:
+                    if new_fitness <= old_fitness:
                         stagnation_count += 1
                         # Check if stagnation occured
                         if stagnation_count > max_stagnation_steps:
@@ -184,12 +192,13 @@ class GeneticAlgorithmTemplate:
                                 if drop_save:
                                     print("Saving fittest candidate.")
                                     os.makedirs(drop_save, exist_ok=True)
-                                    new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
+                                    self.candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5',
+                                                          key='data')
                                     with h5py.File(f'{drop_save}/PopulationDrop_{self.drop_count}.h5') as file:
                                         file['target'] = target
 
                                 self.drop_count += 1
-                                self.winner = pd.DataFrame()
+                                self.candidate = pd.DataFrame()
 
                                 if new_pop_on_drop:
                                     print("Creating new population.")
@@ -197,12 +206,11 @@ class GeneticAlgorithmTemplate:
                                     continue
                                 else:
                                     print("Dropping candidate from population!")
-                                    self.current_winners = self.current_winners.drop(new_candidate.index)
-                                    self.pop = self.pop.drop(new_candidate.index)
+                                    self.current_winners = self.current_winners.drop(self.candidate.index)
                             else:
                                 print("Returning fittest member!")
                                 print("")
-                                return new_candidate
+                                return self.winner
                     else:
                         # Reset stagnation counter
                         stagnation_count = 0
@@ -217,10 +225,7 @@ class GeneticAlgorithmTemplate:
 
             # Update current winning genes
             ##############################
-            if self.winner.empty:
-                self.winner = self.candidate
-            # Cast floats since truth value of a pd.Series is ambiguous
-            elif float(self.candidate['fitness']) > float(self.winner['fitness']):
+            if self.winner.empty or float(self.candidate['fitness']) > float(self.winner['fitness']):
                 self.winner = self.candidate
                 print('Fittest gene in current population is also the globally fittest gene.')
             else:
@@ -240,9 +245,10 @@ class GeneticAlgorithmTemplate:
                 print("Minimum fitness criterion reached!")
                 if enforce_max_iter:
                     if drop_save:
-                        print("Saving candidate!")
-                        new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
+                        print("Saving winner!")
+                        self.winner.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                     self.drop_count += 1
+                    stagnation_count = 0
                     self.winner = pd.DataFrame()
                     if new_pop_on_drop:
                         print(f'Generating new population')
@@ -250,10 +256,11 @@ class GeneticAlgorithmTemplate:
                         continue
                     else:
                         print("Dropping candidate from population!")
-                        self.pop = self.pop.drop(self.candidate.index)
+                        self.pop = self.pop.drop(new_candidate.index)
                         self.current_winners = self.current_winners.drop(new_candidate.index)
+                        self.candidate = pd.DataFrame()
                 else:
-                    return self.candidate
+                    return new_candidate
 
             # Create offspring from current population
             ##########################################
@@ -382,7 +389,7 @@ class GeneticAlgorithmTemplate:
                 if self.pop.at[winner, 'fitness'] > self.current_winners.at[idx_old, 'fitness'] and not \
                         (self.current_winners.loc[:, self.gene_names] == self.pop.loc[winner, self.gene_names]
                          ).all(1).any():
-                    self.current_winners.loc[idx_old, :] = self.pop.drop(index=winner).iloc[0, :]
+                    self.current_winners.loc[idx_old, :] = self.pop.loc[winner, :]
         else:
             self.current_winners = self.pop.nlargest(n_winners, 'fitness')
         return [(self.current_winners.loc[i, self.gene_names], self.current_winners.at[i, 'sigma'])
@@ -393,14 +400,17 @@ class GeneticAlgorithmTemplate:
         mu_new = []
         sigma_new = []
         for i, (mu, sigma) in enumerate(zip(parent[0], parent[1])):
-            mu_temp = np.random.normal(mu, sigma)
+            mu_tmp = np.random.normal(mu, sigma)
+            sigma_tmp = sigma
             j = 0
-            while any([mu_temp < self.initial_gene_pool[self.gene_names[i]]['min'],
-                       mu_temp > self.initial_gene_pool[self.gene_names[i]]['max']]) and j < max_iter:
-                mu_temp = np.random.normal(mu, sigma)
-                sigma *= 0.99
+            while any([mu_tmp < self.initial_gene_pool[self.gene_names[i]]['min'],
+                       mu_tmp > self.initial_gene_pool[self.gene_names[i]]['max']]) and j < max_iter:
+                mu_tmp = np.random.normal(mu, sigma_tmp)
+                sigma_tmp *= 0.99
                 j += 1
-            mu_new.append(mu_temp if j < max_iter else mu)
+            if j > 1:
+                sigma = sigma_tmp
+            mu_new.append(mu_tmp if j < max_iter else mu)
             # Adapt sigma (Beyer1995, p.5)
             xi = np.exp(self.sigma_adapt*np.random.randn())
             sigma_new.append(sigma*xi)
@@ -533,6 +543,176 @@ class GeneticAlgorithmTemplate:
         raise NotImplementedError
 
 
+class DifferentialEvolutionAlgorithm(GeneticAlgorithmTemplate):
+
+    def __init__(self, seed=1234, max_models=1e6):
+        super().__init__()
+        np.random.seed(seed)
+        self.model_ids = list(np.arange(start=0, stop=max_models, step=1, dtype=np.int32)[::-1])
+
+    def run(self, initial_gene_pool: dict, gene_map: dict, loss_func: callable,
+            loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
+            run_kwargs: Optional[dict] = None, template: Optional[Union[str, CircuitTemplate]] = None,
+            compile_kwargs: Optional[dict] = None, save_dir: Optional[str] = "", verbose: bool = True,
+            supress_runtime_warnings: bool = True, **kwargs):
+        """Run a genetic algorithm to optimize genes of a population with respect to a given loss function.
+
+        Parameters
+        ----------
+        initial_gene_pool
+            Dictionary containing ranges for each gene to sample from
+        gene_map
+            Dictionary that provides pointer to model variables for each gene specified in `initial_gene_pool`.
+        loss_func
+            Function that will be used to evaluate the non-fitness of a certain model candidate.
+        loss_kwargs
+            Additional keyword arguments that will be passed onto `loss_func`.
+        run_func
+            Function that will be used to calculate the behavior of a model candidate. The return value of this function
+            will be passed onto `loss_func` as first argument. If not provided, you need to specify a `template`,
+            the `compile_kwargs` and the `run_func_kwargs`. Those will be used to call `CircuitTemplate.from_yaml`,
+            'CircuitIR.compile' and `CircuitIR.run`, respectively.
+        run_kwargs
+            Additional keyword arguments that will be passed onto `run_func`, if `run_func` is provided. Else, they
+            will be passed onto `CircuitIR.run`.
+        template
+            Either a path to a yaml definition of a `CircuitTemplate` or a `pyrates.frontend.CircuitTemplate` instance.
+            Only required, if no `run_func` is provided.
+        compile_kwargs
+            Additional keyword arguments that will be passed onto `CircuitIR.compile`. Only required if no `run_func`
+            is specified.
+        verbose
+            If true, status updates will be printed.
+        supress_runtime_warnings
+            If true, runtime warnings are suppressed.
+        kwargs
+            Additional keyword arguments that will be passed onto `scipy.optimize.differential_evolution`.
+
+        Returns
+        -------
+        pandas.DataFrame containing the overall fittest member of all computed populations
+
+        """
+
+        import h5py
+        import warnings
+
+        self.initial_gene_pool = initial_gene_pool
+        self.num_genes = len(initial_gene_pool)
+
+        # supress warnings
+        if supress_runtime_warnings:
+            warnings.filterwarnings("ignore", category=RuntimeWarning)
+
+        # prepare arguments for call to scipy.optimize.differential_evolution
+        if verbose:
+            print('Starting preparations of evolutionary model optimization.')
+        genes = list(initial_gene_pool)
+        gene_bounds = [(initial_gene_pool[g]['min'], initial_gene_pool[g]['max']) for g in genes]
+        gene_map = [gene_map[g] for g in genes]
+
+        # perform differential evolution optimization
+        if verbose:
+            print('Starting evolutionary optimization.')
+        results = differential_evolution(self.eval_fitness, bounds=gene_bounds,
+                                         args=(gene_map, loss_func, loss_kwargs, run_func, template,
+                                               compile_kwargs, run_kwargs),
+                                         **kwargs)
+
+        # extract results
+        final_genes = results.x
+        final_loss = results.fun
+        optim_sucess = results.success
+
+        # print results
+        if verbose:
+            print('Finished.')
+            print('\n')
+            if optim_sucess:
+                print('Successfully finished the evolutionary optimization of the model.')
+            else:
+                print('Failed to finish the evolutionary optimization of the model successfully.')
+                print(f'Termination message: \n {results.message}')
+            print(f'Final value of the loss function: \n {final_loss}')
+            print(f'Final gene set: \n {final_genes}')
+
+        # save results
+        self.winner = pd.DataFrame(columns=genes + ['loss'], data=[list(final_genes) + [final_loss]])
+        if save_dir:
+            path = f'{save_dir}/fittest_candidate.h5'
+            print(f"Saving final gene set to: {path}")
+            self.winner.to_hdf(path, key='data')
+
+        # remove temporary data
+        try:
+            from shutil import rmtree
+            rmtree('pyrates_build')
+        except FileNotFoundError:
+            pass
+
+        return self.winner
+
+    def eval_fitness(self, genes: list, gene_map: list, loss_func: callable,
+                     loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
+                     template: Optional[str] = None, compile_kwargs: Optional[dict] = None,
+                     run_kwargs: Optional[dict] = None) -> float:
+        """Performs simulation in PyRates of the model defined by `template` and calculates the fitness based on the
+        resulting timeseries of that simulation.
+
+        Parameters
+        ----------
+        genes
+            List of model parameters
+        gene_map
+            List of string-based variable pointers that indicate which gene refers to which variable in the model.
+        loss_func
+        loss_kwargs
+        run_func
+            User-specified run function that can be specified instead of a template, to run user-specified routines.
+        template
+        compile_kwargs
+        run_kwargs
+
+        Returns
+        -------
+
+        """
+
+        if run_func is None:
+
+            attempts = 1
+            while attempts < 100:
+
+                try:
+
+                    # load model template
+                    model_id = self.model_ids.pop()
+                    if type(template) is str:
+                        template = CircuitTemplate.from_yaml(template)
+                    model = deepcopy(template).apply(label=f'model_{model_id}')
+
+                    # apply new parameters to model template
+                    params, param_map = dict(), dict()
+                    for i, (p, key) in enumerate(zip(genes, gene_map)):
+                        params[i] = p
+                        param_map[i] = key
+                    adapt_circuit(model, params=params, param_map=param_map)
+
+                    # compile model into backend
+                    model_compiled = model.compile(**compile_kwargs)
+
+                    # define run func
+                    run_func = model_compiled.run
+                    break
+
+                except (FileExistsError, FileNotFoundError):
+                    attempts += 1
+                    continue
+
+        results = run_func(**run_kwargs)
+        return loss_func(results, **loss_kwargs)
+
+
 class GSGeneticAlgorithm(GeneticAlgorithmTemplate):
     from scipy.spatial.distance import cdist
 
@@ -603,6 +783,225 @@ class CGSGeneticAlgorithm(GeneticAlgorithmTemplate):
             self.pop.at[i, 'fitness'] = float(results.loc['fitness', tuple(candidate_genes)])
 
 
+class ClusterDiffEvAlgorithm(DifferentialEvolutionAlgorithm, ClusterCompute):
+
+    def __init__(self, nodes: list, compute_dir: str, verbose: bool = True):
+        super(DifferentialEvolutionAlgorithm).__init__()
+        super(ClusterCompute).__init__(nodes, compute_dir=compute_dir, verbose=verbose)
+
+    def run(self, initial_gene_pool: dict, gene_map: dict, loss_func: callable,
+            loss_kwargs: Optional[dict] = None, run_func: Optional[callable] = None,
+            run_kwargs: Optional[dict] = None, template: Optional[Union[str, CircuitTemplate]] = None,
+            compile_kwargs: Optional[dict] = None, verbose: bool = True, time_limit: Optional[float] = None,
+            worker_env: Optional[Union[str, list]] = sys.executable, worker_save_dir: Optional[str] = "", **kwargs):
+
+        import h5py
+
+        if verbose:
+            sys.stdout = StreamTee(sys.stdout, self.global_logfile)
+            sys.stderr = StreamTee(sys.stderr, self.global_logfile)
+        else:
+            sys.stdout = open(self.global_logfile, 'a')
+            sys.stderr = StreamTee(sys.stderr, self.global_logfile)
+
+        t_total = t.time()
+
+        # Add local environment to each worker
+        ######################################
+
+        for i, client in enumerate(self.clients):
+            if isinstance(worker_env, list):
+                client['worker_env'] = worker_env[i]
+            else:
+                client['worker_env'] = worker_env
+
+        # Create global result file
+        ############################
+
+        if verbose:
+            print("***Creating global results file***")
+        t0 = t.time()
+
+        # Create result directory and result file for current parameter grid
+        global_res_file = f'{self.compute_dir}/optimization_result.h5'
+
+        # Write gene pool and config information to global result file
+        with h5py.File(global_res_file, 'a') as file:
+            for key, value in initial_gene_pool.items():
+                file.create_dataset(f'/GenePool/Keys/{key}', data=value)
+        # TODO: add global configuration details to file
+        if verbose:
+            print(f'Done. Elapsed time: {t.time() - t0:.3f} seconds \n\n')
+
+        # Create keyword dictionary for threads
+        #######################################
+        thread_kwargs = {
+            "worker_env": worker_env,
+            "initial_gene_pool": initial_gene_pool,
+            "gene_map": gene_map,
+            "loss_func": loss_func,
+            "loss_kwargs": loss_kwargs,
+            "run_func": run_func,
+            "run_kwargs": run_kwargs,
+            "compile_kwargs": compile_kwargs,
+            "template": template,
+            "save_dir": worker_save_dir
+        }
+
+        # Start cluster computation
+        ###########################
+
+        if verbose:
+            print("***Starting cluster computation***")
+
+        # Spawn threads to control each node connection and start computation
+        threads = [self.spawn_thread(client, thread_kwargs, timeout=time_limit) for client in self.clients]
+        # Wait for all threads to finish
+        for t_ in threads:
+            t_.join()
+
+        print("")
+        print(f'Cluster computation finished. Elapsed time: {t.time() - t_total:.3f} seconds')
+
+        # Write local results to global result file
+        ###########################################
+
+        print(f'***WRITING RESULTS TO GLOBAL RESULT FILE***')
+        t0 = t.time()
+
+        # Get sorted list of temporary result files to iterate through
+        temp_res_files = glob.glob(f'{self.compute_dir}/*_temp*')
+        temp_res_files.sort()
+
+        # Read number of different circuit outputs and prepare lists to concatenate results
+        res_dict = {}
+        try:
+            with pd.HDFStore(temp_res_files[0], "r") as store:
+                for key in store.keys():
+                    res_dict[key] = []
+        except IndexError:
+            print("No result file created. Check local log files for worker script errors.")
+            return ""
+
+        # Concatenate results from each temporary result file
+        for file in temp_res_files:
+            with pd.HDFStore(file, "r") as store:
+                for idx, key in enumerate(store.keys()):
+                    res_dict[key].append(store[key])
+
+        # Create DataFrame for each output variable and write to global result file
+        with pd.HDFStore(global_res_file, "a") as store:
+            for key, value in res_dict.items():
+                if key != '/result_map' and len(value) > 0:
+                    df = pd.concat(value, axis=kwargs.pop('result_concat_axis', 1))
+                    store.put(key=f'/Results{key}', value=df)
+            result_map = pd.concat(res_dict['/result_map'], axis=0)
+            store.put(key=f'/Results/result_map', value=result_map)
+
+        # Delete temporary local result files
+        for file in temp_res_files:
+            os.remove(file)
+
+        print(f'Elapsed time: {t.time() - t0:.3f} seconds')
+        print(f'Find results in: {self.compute_dir}/optimization_results.h5')
+        print("")
+
+        return global_res_file
+
+    def thread_master(self, client, thread_kwargs: dict, timeout: float):
+        """
+
+        Parameters
+        ----------
+        client
+        thread_kwargs
+        timeout
+
+        Returns
+        -------
+
+        """
+
+        from threading import currentThread
+
+        # This lock ensures that parameter chunks are fetched by workers in the same order as they are defined
+        # in the node list
+        self.lock.acquire()
+
+        thread_name = currentThread().getName()
+
+        # Get client information
+        pm_client = client["paramiko_client"]
+        logfile = client["logfile"]
+        worker_env = client["worker_env"]
+
+        # Get keyword arguments
+        config_file = thread_kwargs["config_file"]
+        worker_file = thread_kwargs["worker_file"]
+
+        # Prepare worker command
+        command = f'{worker_env} {worker_file}'
+
+        # Create build_dir for grid_search numpy backend
+        local_build_dir = f'{self.compute_dir}/{thread_name}'
+        os.makedirs(local_build_dir, exist_ok=True)
+
+        # Start scheduler
+        #################
+
+        # Disable thread switching
+        self.lock.acquire()
+
+        # Execute worker script on the remote host
+        print(f'[T]\'{thread_name}\': Starting remote computation...')
+
+        channel = pm_client.get_transport().open_session()
+        channel.settimeout(timeout)
+
+        # Execute the given command
+        channel.get_pty()
+        channel.exec_command(command +
+                             f' --config_file={config_file}'
+                             f' --local_res_file={local_build_dir}'
+                             f' --build_dir={local_build_dir}'
+                             f' &>{logfile}',  # redirect and append stdout
+                             )
+
+        # Enable thread switching
+        self.lock.release()
+
+        # Release the second lock if its still acquired from the very beginning
+        try:
+            self.lock.release()
+        except RuntimeError:
+            pass
+
+        # Wait for remote computation exit status
+        # (independent of success or failure)
+        #########################################
+
+        nbytes = 1024
+        t1 = t.time()
+        while not channel.exit_status_ready():
+            if channel.recv_ready():
+                data = channel.recv(nbytes)
+                while data:
+                    data = channel.recv(nbytes)
+            if channel.recv_stderr_ready():
+                error_buff = channel.recv_stderr(nbytes)
+                while error_buff:
+                    error_buff = channel.recv_stderr(nbytes)
+            t2 = t.time()
+            if t2 - t1 > timeout:
+                channel.close()
+                break
+            t.sleep(1.0)
+        channel.close()
+        exit_status = channel.recv_exit_status()
+
+        return exit_status
+
+
 def plot_results_2d(p1: str, p2: str, fname_identifier: str, fname_type: str = '.h5', fitness_measure: str = 'fitness',
                     cbar_kwargs: dict = None, **kwargs):
     """
@@ -651,4 +1050,4 @@ def plot_results_2d(p1: str, p2: str, fname_identifier: str, fname_type: str = '
     cbar_ax = grid.fig.add_axes(pos)  # x, y, width, height
     colorbar(cax=cbar_ax)
 
-    return grid
+    return grid
\ No newline at end of file
